;;; -*- Mode: Lisp -*-
;;; $Id: context.lisp,v 1.80 2003/01/02 17:32:56 jesse Exp $
;;;
;;; Transaction Object and Editing Context support for standard-db-object

(in-package :maisql-sys)

;; odcl::transacted-store interface
;;
;; see odcl/transaction.lisp


(defvar *full-transaction-body* t
  "If true, then ODCL::ATTACH-TRANSACTION for database derived stores
will start a database transaction, effictively putting the entire body
of WITH-CONTEXT inside a database level transaction.")

(defmethod odcl::attach-transaction ((self database) attach-transaction)
  (call-next-method)
  ;; after attaching, start the database level transaction?
  (when *full-transaction-body*
    (database-start-transaction self)))

(defmethod odcl::detach-transaction ((self database) detach-transaction)
  ;; only abort if we have a database transaction in play, this avoids
  ;; a warning being printed to console
  (call-next-method)
  (when (< 0 (slot-value self 'transaction-level))
    (database-abort-transaction self)))


;; Since databases change their class when closed, we need to implement
;; part of the transactable-store interface for closed-database, even tho
;; it doesn't have a transactable-store mixin.
(defmethod odcl::attached-to-transaction? ((self closed-database) transaction)
  (declare (ignore transaction))
  nil)

;; Inside a with-context, make ourselves the *default-database*
(defmethod odcl:editing-context-bindings ((store database))
  (list (cons '*default-database* store)))

(defmethod odcl::commit-phase-1 ((self database) transaction)
  (without-context
   (let ((stack (remove-if-not (lambda (event)
                                 (eql self (odcl::tx-event-store event)))
                               (slot-value transaction 'odcl::stack))))
     (let ((completed nil))
       (unwind-protect
            (progn
              ;; we do not want to increment the transaction-level
              ;; if attach-transaction  already started it 
              (when (and (not *full-transaction-body*)
                         (= 0 (slot-value self 'transaction-level)))
                (database-start-transaction self))
              (dolist (action (reverse stack))
                ;; (cmsg "indexed-store cp1: ~s ~s" self action)
                (let ((data (odcl::tx-event-data action)))
                  (ecase (odcl::tx-event-type action)
                    (:rollback
                     t)
                    ((:edit :insert)
                     (update-records-from-instance data))
                    (:delete
                     (delete-instance-records data)))))
              (setf completed t))
         (unless completed
           (database-abort-transaction self)))
       (lambda (action)
         (ecase action
           (:commit-phase-2
            (database-commit-transaction self))
           (:rollback
            (database-abort-transaction self))))))))

(defmethod odcl::generate-new-oid ((self database) class)
  (sequence-next (sequence-from-class class)))

(defmethod odcl::generate-copy ((self standard-db-object) new-oid)
  (let* ((*db-initializing* t)
         (copy-of-self (make-instance (type-of self))))
    (dolist (slot (pcl::class-slots (find-class (type-of self))))
      (when (member (view-class-slot-db-kind slot) '(:key :base))
        (let ((slot-name (slot-value slot 'pcl::name)))
          (setf (slot-value copy-of-self slot-name)
                (slot-value self slot-name)))))
    (setf (slot-value copy-of-self 'oid) new-oid)
    copy-of-self))
  
(defun instance-key-to-qualifier (key)
  (destructuring-bind (class-name &rest key-values)
      key
    (when-bind (class (find-class class-name nil))
      (when (typep class 'view-metaclass)
        (let ((qual-list (mapcar (lambda (key value)
                                   (unless key
                                     (error "mismatched key count"))
                                   (unless value
                                     (error "missing key value"))
                                   (let ((slot-name (slot-definition-name key)))
                                     (sql-operation '==
                                                    (sql-slot-value class-name slot-name)
                                                    (sql-slot-value class-name slot-name value))))
                                 (slot-value class 'key-slots)
                                 key-values)))
          (if (< 1 (length qual-list))
              (apply #'sql-and qual-list)
              (car qual-list)))))))
          
(defmethod get-store-instance-by-key ((database database) key)
  (when-bind (qualifier (instance-key-to-qualifier key))
    (car (select (car key) :where qualifier))))

;; odcl::transaction-object interface
;;
;; see odcl/transaction.lisp

(defmethod instance-key ((instance standard-db-object))
  (cons (type-of instance)
        (mapcar (lambda (slotdef)
                  (let ((name (slot-definition-name slotdef)))
                    (unless (slot-boundp instance name)
                      (return-from instance-key nil))
                    (slot-value instance name)))
                (key-slots (class-of instance)))))

#-lispworks
(defmethod instance-snapshot ((instance standard-db-object) &aux ss)
  (let ((*db-deserializing* t)
        (unbound (gensym)))
    (dolist (slot (pcl:class-slots (pcl:class-of instance)))
      (let ((name (slot-value slot 'pcl::name)))
        (push (cons name (if (slot-boundp instance name)
                             (slot-value instance name)
                             unbound)) ss)))
    (cons :snapshot (cons unbound ss))))

;; lispworks patch submitted by Martin Thornquist <martint@ifi.uio.no>
#+lispworks
(defmethod instance-snapshot ((instance standard-db-object) &aux ss)
  (let ((*db-deserializing* t)
        (unbound (gensym)))
    (dolist (slot (class-slots (class-of instance)))
      (let ((name (slot-definition-name slot)))
        (push (cons name (if (slot-boundp instance name)
                             (slot-value instance name)
                             unbound)) ss)))
    (cons :snapshot (cons unbound ss))))

(defmethod instance-snapshot ((instance standard-db-object) &aux ss)
  (let ((*db-deserializing* t)
        (unbound (gensym)))
    (dolist (slot (pcl:class-slots (pcl:class-of instance)))
      (let ((name (slot-value slot 'pcl::name)))
        (push (cons name (if (slot-boundp instance name)
                             (slot-value instance name)
                             unbound)) ss)))
    (cons :snapshot (cons unbound ss))))

(defmethod instance-restore ((instance standard-db-object) ss)
  (destructuring-bind (snapshot-tag unbound-marker &rest slots)
      ss
    (if (not (eql snapshot-tag :snapshot))
        (error "invalid snapshot")
        (let ((*db-deserializing* t))
          (dolist (slot slots)
            (if (eq unbound-marker (cdr slot))
                (slot-makunbound instance (car slot))
                (setf (slot-value instance (car slot))
                      (cdr slot))))))))
