(in-package :maisql-sys)

;;
;; Functions for schema manipulation
;;

(defvar *table-schemas* (make-hash-table :test #'equal)
  "Hash of schema name to table lists.")

(defun create-table (name description &key (database *default-database*)
                      (constraints nil))
  "Create a table with the given NAME, containing the attributes in
DESCRIPTION."
  (let ((stmt (make-instance 'sql-create-table
			     :name name
			     :columns description
			     :modifiers constraints)))
    (execute-command stmt :database database)
    (pushnew name (gethash *default-schema* *table-schemas*) :test #'equal)))




(defun drop-table (name &key (if-does-not-exist :error) (database *default-database*) &aux table-name)
  "Drops TABLE from *default-database*, or if a :database keyword
argument is given, the specified database."
  (etypecase name
    (string
     (setf table-name name))
    (sql-ident
     (setf table-name (generate-sql name))))
  (ecase if-does-not-exist
    (:ignore
     (unless (tablep table-name :database database)
       (return-from drop-table nil)))
    (:error
     t))
  (let ((expr (concatenate 'string "DROP TABLE " table-name)))
    (execute-command expr :database database)
    t))

(defun list-tables (&key (database *default-database*)
			 (system-tables nil))
  "List all tables i*default-database, or if the :database keyword arg
is given, the specified database.  If the keyword arg :system-tables
is true, then it will not filter out non-user tables.  Table names are
given back as a list of strings."
  (database-list-tables database :system-tables system-tables))

(defun tablep (table &key (database *default-database*))
  (database-tablep database table))

(defvar sql-sys::*index-schemas* (make-hash-table :test #'equal)
  "Hash of schema name to index lists.")


(defun create-index (idxname table columns &key (database *default-database*) (unique nil))
  (let ((columns (mapcar #'(lambda (name)
                             (etypecase name
                               (string (string-upcase name))
                               (sql-sys::sql-ident
                                (sql name))
                               (symbol
                                (sql-escape (sql name)))))
                         (listify columns))))
    (execute-command (format nil "create ~a index ~a on ~a (~{~a~^, ~})"
			     (if unique "unique" "")
                             (sql-escape idxname) (sql-escape table) columns)
                     :database database)
    (pushnew idxname (gethash sql-sys::*default-schema* sql-sys::*index-schemas*))))

(defun drop-index (idxname &key (database *default-database*) (if-does-not-exist :error))
  (case if-does-not-exist
    (:ignore
     (unless (indexp idxname)
       (return-from drop-index)))
    (t
     (execute-command (format nil "drop index ~a" idxname) :database database))))


(defun list-indexes (&key (database *default-database*)
                       (system-indexes nil))
  "List all sequences in *default-database*, or if the :database
keyword arg is given, the specified database.  If the keyword arg
:system-tables is true, then it will not filter out non-user tables.
Table names are given back as a list of strings."
  (database-list-indexes database :system-indexes system-indexes))

(defun indexp (index &key (database *default-database*))
  (database-indexp database index))

(defun list-attributes (table &key (database *default-database*))
  "List the attributes of TABLE in *default-database, or if the
:database keyword is given, the specified database.  Attributes are
returned as a list of strings."
  (database-list-attributes table database))

(defun attribute-type (attribute table &key (database *default-database*))
  "Return the field type of the ATTRIBUTE in TABLE.  The optional
keyword argument :database specifies the database to query, defaulting
to *default-database*."
  (database-attribute-type attribute table database))

(defun add-attribute (table attribute &key (database *default-database*))
  "Add the ATTRIBUTE to TABLE.  The ATTRIBUTE sepcification must
include a type argument.  The optional keyword argument :database
specifies the database to operation on, defaulting to
*default-database*."
  (database-add-attribute table attribute database))

(defun rename-attribute (table oldatt newname
			       &key (database *default-database*))
  (error "(rename-attribute ~a ~a ~a ~a) is not implemented" table oldatt newname database))
